import typing

from fastapi import Response as FResponse, FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import StreamingResponse as FStreamingResponse
from nestipy.common.http_ import Response
from nestipy.types_ import CallableHandler, MountHandler, WebsocketHandler
from starlette.middleware import _MiddlewareClass
from starlette.types import ASGIApp

from .http_adapter import HttpAdapter


class FastApiAdapter(HttpAdapter):
    def __init__(self):
        self.instance = FastAPI(
            on_startup=[self.on_startup],
            on_shutdown=[self.on_shutdown],
            openapi_url=None,
        )

    def get_instance(self) -> any:
        return self.instance

    def create_wilchard(self, prefix: str = '/', name: str = 'full_path') -> str:
        return f"/{prefix.strip('/')}" + "/{" + f"{name}:path" + "}"

    def use(self, callback: CallableHandler, metadata: dict) -> None:
        # need to transform if we use middleware from here
        self.instance.middleware('http')

    def get(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        self.instance.get(route)(self._create_fastapi_handler(callback, metadata))

    def post(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        self.instance.post(route)(self._create_fastapi_handler(callback, metadata))

    def ws(self, route: str, callback: WebsocketHandler, metadata: dict) -> None:
        pass

    def mount(self, route: str, callback: MountHandler) -> None:
        self.instance.mount(route, typing.cast(ASGIApp, callback))

    def put(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        self.instance.put(route)(self._create_fastapi_handler(callback, metadata))

    def delete(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        self.instance.delete(route)(self._create_fastapi_handler(callback, metadata))

    def patch(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        self.instance.patch(route)(self._create_fastapi_handler(callback, metadata))

    def options(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        self.instance.options(route)(self._create_fastapi_handler(callback, metadata))

    def head(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        self.instance.head(route)(self._create_fastapi_handler(callback, metadata))

    def all(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        pass

    def engine(self, args, *kwargs) -> None:
        pass

    def enable_cors(self) -> None:
        self.instance.add_middleware(
            typing.cast(
                _MiddlewareClass[typing.Any],
                CORSMiddleware
            ),
            allow_origins=['*'],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

    def _create_fastapi_handler(self, callback: CallableHandler, metadata: dict):
        # path = metadata['path']
        # params = RouteParamsExtractor.extract_params(path)
        async def fastapi_handler() -> FResponse:
            req, res, next_fn = self.create_handler_parameter()
            result: Response = await callback(req, res, next_fn)
            if result.is_stream():
                return FStreamingResponse(
                    content=result.get_stream(),
                    headers={k: v for k, v in result.headers()},
                    status_code=result.status_code() or 200
                )
            return FResponse(
                content=result.content(),
                headers={k: v for k, v in result.headers()},
                status_code=result.status_code() or 200
            )

        return fastapi_handler



import dataclasses
import inspect
from functools import lru_cache
from typing import Type, Union, Any, Optional, Callable, Awaitable, TYPE_CHECKING

from nestipy.metadata import ClassMetadata, CtxDepKey, ModuleMetadata, ProviderToken, Reflect
from pydantic import BaseModel

from .context_container import RequestContextContainer
from .dependency import TypeAnnotated
from .meta import ContainerHelper
from .utils import uniq

if TYPE_CHECKING:
    from .provider import ModuleProviderDict


class NestipyContainer:
    _instance: "NestipyContainer" = None
    _services = {}
    _global_service_instances = {}
    _singleton_instances = {}
    _singleton_classes = set()
    helper = ContainerHelper()

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(NestipyContainer, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    @classmethod
    def get_instance(cls, *args, **kwargs):
        return NestipyContainer(*args, **kwargs)

    def add_transient(self, service: Type):
        self._services[service] = service

    def add_singleton(self, service: Type):
        self._services[service] = service
        self._singleton_classes.add(service)

    def get_all_services(self) -> list:
        return list(self._services.keys())

    def add_singleton_instance(self, service: Union[Type, str], service_instance: object):
        self._singleton_instances[service] = service_instance

    def get_all_singleton_instance(self) -> list:
        return [v for k, v in self._singleton_instances.items()]

    @classmethod
    @lru_cache()
    def get_global_providers(cls) -> list:
        global_providers = []
        for service in cls._services:
            if service in global_providers:
                continue
            metadata: ClassMetadata = Reflect.get_metadata(service, ClassMetadata.Metadata, None)
            if metadata is not None:
                is_global = Reflect.get_metadata(metadata.get_module(), ModuleMetadata.Global, False)
                is_root = Reflect.get_metadata(metadata.get_module(), ModuleMetadata.Root, False)
                if is_global or is_root:
                    global_providers += Reflect.get_metadata(metadata.get_module(), ModuleMetadata.Providers, [])
        return uniq(global_providers)

    @classmethod
    def get_dependency_metadata(cls, service: Union[Type, object]) -> list:
        from .provider import ModuleProviderDict
        # extract global data from _service, not from module because all provider is already saved in _services of
        # container
        metadata: ClassMetadata = Reflect.get_metadata(service, ClassMetadata.Metadata, None)
        if metadata is not None:
            global_providers = cls.get_global_providers()
            providers, import_providers = metadata.get_service_providers()
            uniq_providers = []
            for m in uniq(providers + global_providers + import_providers):
                if isinstance(m, ModuleProviderDict):
                    uniq_providers.append(m.token)
                else:
                    uniq_providers.append(m)
            return uniq(uniq_providers)
        # raise ValueError(f"Dependency Metadata not found  for {service.__name__} service ")
        return []

    @classmethod
    def _data_to_typed(cls, annotation: Union[Type, Any], data: Union[dict, list, str, int, tuple, set, Any]):
        if dataclasses.is_dataclass(annotation) or issubclass(annotation, BaseModel):
            return annotation(**data)
        elif annotation in (dict, list, str, int, tuple, set, Any):
            return data
        return None

    @classmethod
    async def _resolve_context_service(cls, dep_key: TypeAnnotated, annotation: Union[Type, Any]):
        context_container = RequestContextContainer.get_instance()
        callback = dep_key.metadata.callback
        if inspect.iscoroutinefunction(callback):
            return await callback(dep_key.metadata.token, annotation, context_container)
        else:
            return callback(dep_key.metadata.token, annotation, context_container)

    async def _resolve_module_provider_dict(self, instance: "ModuleProviderDict", search_scope: list):
        if instance.value:
            return instance.value
        elif instance.existing:
            if isinstance(instance.existing, ProviderToken):
                return await self.get(instance.existing.key)
            else:
                return await self.get(instance.existing)
        elif instance.use_class:
            return await self.get(instance.use_class)
        elif instance.factory:
            return await self.resolve_factory(
                factory=instance.factory,
                inject=instance.inject,
                search_scope=search_scope
            )

        else:
            return None

    async def _check_exist_singleton(self, key: Union[Type, str]):
        from .provider import ModuleProviderDict
        if key in self._singleton_instances:
            instance = self._singleton_instances[key]
            # to keep improve
            if isinstance(instance, ModuleProviderDict):
                search_scope = self.get_dependency_metadata(instance)
                if instance.token in search_scope:
                    value = await self._resolve_module_provider_dict(instance, search_scope=search_scope)
                    # update singleton instance to have the async valur from ModuleProviderDict
                    self._singleton_instances[key] = value
                    return value
                else:
                    raise ValueError(
                        f"Service {instance.__class__.__name__} "
                        f"not found in scope")
            else:
                return instance
        return None

    def _check_service(self, key: Union[Type, str], origin: Optional[list] = None) -> tuple:
        if key not in self._services:
            raise ValueError(f"Service {key} not found")
        service = self._services[key]
        if service in (origin or []):
            raise ValueError(f"Circular dependency found  for {service.__name__} service ")
        return service, origin or set()

    async def _resolve_property(
            self,
            key: Union[Type, str],
            origin: Optional[list] = None,
            disable_scope: bool = False
    ):
        service, origin = self._check_service(key, origin)
        search_scope = self.get_dependency_metadata(service)
        origin.add(service)
        annotations: dict = getattr(service, '__annotations__', {})
        for name, param_annotation in annotations.items():
            annotation, dep_key = self.helper.get_type_from_annotation(param_annotation)
            if dep_key.metadata.key in CtxDepKey.to_list():
                if dep_key.metadata.key is not CtxDepKey.Service:
                    dependency = await self._resolve_context_service(dep_key, annotation)
                    setattr(service, name, dependency)
                elif dep_key.metadata.token in search_scope or annotation in search_scope or disable_scope:
                    dependency = await self.get(dep_key.metadata.token or annotation)
                    setattr(service, name, dependency)
                else:
                    _name: str = annotation.__name__ if not isinstance(annotation, str) else annotation
                    raise ValueError(
                        f"Service {_name} "
                        f"not found in scope of {service.__name__}")
            else:
                continue
        origin.remove(service)
        self._services[key] = service

    async def _get_method_dependency(self, method_to_resolve: Callable, search_scope: list, origin: list):
        params = inspect.signature(method_to_resolve).parameters
        args = {}
        for name, param in params.items():
            if name != 'self' and param.annotation is not inspect.Parameter.empty:
                annotation, dep_key = self.helper.get_type_from_annotation(param.annotation)
                if dep_key.metadata.key in CtxDepKey.to_list() and dep_key.metadata.key is not CtxDepKey.Service:
                    dependency = await self._resolve_context_service(dep_key, annotation)
                    args[name] = dependency
                elif annotation in search_scope:
                    dependency = await self.get(annotation, origin=origin)
                    args[name] = dependency
                else:
                    _name: str = annotation.__name__ if not isinstance(annotation, str) else annotation
                    raise ValueError(f"Service {_name} not found in scope {search_scope}")
        return args

    @classmethod
    async def _call_method(cls, method: Callable, args: dict):
        if inspect.iscoroutinefunction(method):
            return await method(**args)
        return method(**args)

    async def resolve_factory(self, factory: Callable, inject: list, search_scope: list):
        search_scope_by_inject = [m for m in inject if m in search_scope]
        args = await self._get_method_dependency(
            method_to_resolve=factory,
            search_scope=search_scope_by_inject,
            origin=[]
        )
        return await self._call_method(method=factory, args=args)

    async def _resolve_method(
            self,
            key: Union[Type, str, object],
            method: str = '__init__',
            origin: Optional[list] = None
    ):
        service, origin = self._check_service(key, origin)
        search_scope = self.get_dependency_metadata(service)
        origin.add(service)
        method_to_resolve = getattr(service, method, None)
        if not method_to_resolve:
            raise Exception(f"Method {method}  not found in {service.__name__} service ")
        args = await self._get_method_dependency(method_to_resolve, search_scope, origin)
        if method == '__init__':
            result = service(**args)
            if service in self._singleton_classes:
                self._singleton_instances[service] = result
        else:
            # Service must be an instance (controller)
            instance = await self.get(key)
            instance_method = getattr(instance, method, method_to_resolve)
            result = await self._call_method(instance_method, args)

        origin.remove(service)
        return result

    async def get(
            self,
            key: Union[Type, str],
            method: str = '__init__',
            origin: Optional[list] = None,
            disable_scope: Optional[bool] = False
    ) -> Awaitable[object]:
        in_singleton = await self._check_exist_singleton(key=key)
        if in_singleton:
            if method == '__init__':
                return in_singleton
        else:
            await self._resolve_property(key, origin=origin, disable_scope=disable_scope)
        return await self._resolve_method(key, method=method, origin=origin)



from dataclasses import dataclass

from pydantic import create_model

from .decorator import ApiBearerAuth, ApiBasicAuth, ApiSecurity
from .decorator import ApiResponse, ApiCreatedResponse, ApiOkResponse, ApiNotFoundResponse
from .decorator import ApiTags, ApiId, ApiBody, ApiParameter
from .document_builder import DocumentBuilder
from .swagger_module import SwaggerModule


def ApiSchema(cls):
    data_cls = dataclass(cls)
    fields = {}
    for field in data_cls.__dataclass_fields__.values():
        fields[field.name] = (field.type, ...)
    # Schema
    return create_model(data_cls.__name__, **fields)


if __name__ == '__main__':
    # Example usage:
    @ApiSchema
    class TestDto:
        name: str


    @ApiSchema
    class CreateCatDto:
        name: str
        age: int
        breed: str
        test: TestDto

__all__ = [
    'DocumentBuilder',
    'SwaggerModule',
    "ApiResponse",
    "ApiCreatedResponse",
    "ApiOkResponse",
    "ApiNotFoundResponse",
    "ApiTags",
    "ApiId",
    "ApiBody",
    "ApiParameter",
    "ApiBearerAuth",
    "ApiBasicAuth",
    "ApiSecurity"
]



import dataclasses
from base64 import b64decode, b64encode
from typing import Type, Union, Literal

import itsdangerous
import ujson
from itsdangerous.exc import BadSignature

from nestipy.common.decorator import Injectable
from nestipy.common.http_ import Request, Response
from nestipy.common.middleware.interface import NestipyMiddleware
from nestipy.types_ import NextFn


@dataclasses.dataclass
class SessionOption:
    secret_key: str = ''
    session_cookie: str = "session"
    max_age: Union[int, None] = 14 * 24 * 60 * 60
    path: str = "/"
    same_site: Literal["lax", "strict", "none"] = "lax"
    https_only: bool = False
    domain: Union[str, None] = None


def session(option: SessionOption = SessionOption()) -> Type:
    @Injectable()
    class SessionMiddleware(NestipyMiddleware):

        async def use(self, req: Request, res: Response, next_fn: NextFn):
            initial_session_was_empty = True
            signer = itsdangerous.TimestampSigner(str(option.secret_key))
            security_flags = "httponly; samesite=" + option.same_site
            if option.https_only:
                security_flags += "; secure"
            if option.domain is not None:
                security_flags += f"; domain={option.domain}"
            if option.session_cookie in req.cookies:
                data = req.cookies[option.session_cookie].encode("utf-8")
                try:
                    un_sign_data = signer.unsign(data, max_age=option.max_age)
                    req.session = ujson.loads(b64decode(un_sign_data))
                    initial_session_was_empty = False
                except BadSignature:
                    req.session = {}
            else:
                req.session = {}
            result = await next_fn()
            cookie_max_age = f"Max-Age={option.max_age}; " if option.max_age else ""
            header_value = ", ".join([
                f"{key}={value}; path={option.path}; {cookie_max_age}{security_flags}"
                for (key, value) in res.cookies()
            ])
            if header_value:
                header_value += ", "
            if req.session:
                data = b64encode(ujson.dumps(req.session).encode("utf-8"))
                sign_data = signer.sign(data)
                header_value += "{session_cookie}={data}; path={path}; {max_age}{security_flags}".format(  # noqa E501
                    session_cookie=option.session_cookie,
                    data=sign_data.decode("utf-8"),
                    path=option.path,
                    max_age=f"Max-Age={option.max_age}; " if option.max_age else "",
                    security_flags=security_flags,
                )
                res.header('Set-Cookie', header_value)
            elif initial_session_was_empty:
                header_value += "{session_cookie}={data}; path={path}; {expires}{security_flags}".format(  # noqa E501
                    session_cookie=option.session_cookie,
                    data="null",
                    path=option.path,
                    expires="expires=Thu, 01 Jan 1970 00:00:00 GMT; ",
                    security_flags=security_flags,
                )
                res.header("Set-Cookie", header_value)

            return result

    return SessionMiddleware



from typing import Type

from nestipy.metadata import ClassMetadata, ModuleMetadata, Reflect

from .metadata_creator import MetadataCreator


class ModuleMetadataCreator(MetadataCreator):
    def _extract(self) -> list:
        return Reflect.get_metadata(self.module, ModuleMetadata.Imports, [])

    def _type(self) -> Type["MetadataCreator"]:
        return ModuleMetadataCreator

    def create(self) -> None:
        imports = [self.module] + self._extract_import()
        providers = self.extract_providers()
        for im in imports:
            Reflect.set_metadata(
                im,
                ClassMetadata.Metadata,
                ClassMetadata(
                    im,
                    global_providers=providers
                )
            )



# TODO: precise import
from nestipy.ioc import ModuleProviderDict
from .config import ConfigModule, ConfigService
from .decorator import Module, Controller, Injectable, Route, Get, Post, Patch, Put, Delete
from .exception import HttpException, HttpStatusMessages, HttpStatus, UseFilters, Catch, ExceptionKey, ExceptionFilter
from .guards import CanActivate, UseGuards
from .http_ import Request, Response, Websocket, UploadFile
from .interceptor import NestipyInterceptor, UseInterceptors
from .logger import logger
from .middleware import cors, NestipyMiddleware, session, SessionOption
from .template import Render, TemplateEngine

__all__ = [
    "ConfigModule",
    "ConfigService",
    "Module",
    "Controller",
    "Injectable",
    "Route",
    "Get",
    "Post",
    "Put",
    "Patch",
    "Delete",
    "HttpException",
    "HttpStatusMessages",
    "HttpStatus",
    "UseFilters",
    "Catch",
    "ExceptionKey",
    "ExceptionFilter",
    "CanActivate",
    "UseGuards",
    "Response",
    "Request",
    "Websocket",
    "UploadFile",
    "NestipyInterceptor",
    "UseInterceptors",
    "cors",
    "NestipyMiddleware",
    "Render",
    "TemplateEngine",
    "ModuleProviderDict",
    "session",
    "SessionOption",
    "logger"
]



from .builder import DynamicModule, ConfigurableModuleBuilder
from .module import NestipyModule, MiddlewareConsumer

__all__ = [
    "DynamicModule",
    "ConfigurableModuleBuilder",
    "NestipyModule",
    "MiddlewareConsumer"
]



from typing import TYPE_CHECKING

from nestipy.common.decorator import Injectable
from nestipy.common.helpers import SpecialProviderExtractor
from nestipy.common.interceptor import NestipyInterceptor, InterceptorKey
from nestipy.core.constant import APP_INTERCEPTOR
from nestipy.core.context.execution_context import ExecutionContext
from nestipy.ioc import NestipyContainer
from nestipy.metadata import ClassMetadata, Reflect

if TYPE_CHECKING:
    from nestipy.types_.handler import NextFn


@Injectable()
class RequestInterceptor(NestipyInterceptor, SpecialProviderExtractor):
    context: ExecutionContext

    def __init__(self):
        self.container = NestipyContainer.get_instance()

    async def intercept(self, context: ExecutionContext, next_fn: "NextFn"):
        self.context = context
        handler_module_class = context.get_module()
        handler_class = context.get_class()
        handler = context.get_handler()

        global_interceptors = context.get_adapter().get_global_interceptors()
        module_interceptors = self.extract_special_providers(
            handler_module_class,
            NestipyInterceptor,
            APP_INTERCEPTOR
        )
        class_interceptors = Reflect.get_metadata(handler_class, InterceptorKey.Meta, [])
        handler_interceptors = Reflect.get_metadata(handler, InterceptorKey.Meta, [])
        all_interceptors = handler_interceptors + class_interceptors + module_interceptors + global_interceptors
        # setup dependency as the same as the container
        for intercept in all_interceptors:
            if issubclass(intercept, NestipyInterceptor):
                # Put dependency
                services = self.container.get_all_services()
                Reflect.set_metadata(
                    intercept, ClassMetadata.Metadata,
                    ClassMetadata(handler_class, global_providers=services)
                )

        return await self._recursive_aplly_interceptor(0, all_interceptors, next_fn)

    async def _recursive_aplly_interceptor(self, index: int, all_interceptors: list, next_fn: "NextFn"):
        if len(all_interceptors) > index:
            interceptor = all_interceptors[index]
            instance: NestipyInterceptor = await self.container.get(interceptor)

            async def _next_fn():
                return await instance.intercept(self.context, next_fn)

            return await self._recursive_aplly_interceptor(index + 1, all_interceptors, _next_fn)
        else:
            return await next_fn()


__all__ = [
    "RequestInterceptor"
]



class NestipyGraphqlKey:
    resolver: str = '__graphql_resolver__'
    query: str = '__graphql_query__'
    mutation: str = '__graphql_mutation__'
    subscription: str = '__graphql_subscription__'
    handler: str = '__graphql_handler__'
    return_type: str = '__graphql_return_type__'



from typing import Union, Type, Callable

from nestipy.common.decorator import Injectable
from nestipy.metadata import SetMetadata

from .interface import NestipyInterceptor
from .meta import InterceptorKey


def UseInterceptors(*interceptors: Union[Type, NestipyInterceptor]):
    decorator = SetMetadata(InterceptorKey.Meta, list(interceptors), as_list=True)

    def wrapper(cls: Union[Type, Callable]):
        cls = Injectable()(cls)
        return decorator(cls)

    return wrapper



from abc import ABC, abstractmethod
from typing import Tuple, Any, Callable, Union, Type, TYPE_CHECKING

from nestipy.common.exception.http import HttpException
from nestipy.common.http_ import Request, Response, Websocket
from nestipy.common.template import TemplateKey
from nestipy.types_ import CallableHandler, NextFn, WebsocketHandler, MountHandler
from nestipy.websocket.adapter import IoAdapter
from nestipy.metadata import SetMetadata, Reflect

if TYPE_CHECKING:
    from nestipy.common.exception.interface import ExceptionFilter
    from nestipy.common.guards import CanActivate
    from nestipy.common.interceptor import NestipyInterceptor


class HttpAdapter(ABC):
    STATE_KEY: str = '__state__'

    _global_interceptors: list = []
    _global_filters: list = []
    _global_guards: list = []

    startup_hooks: list = []
    shutdown_hook: list = []

    _io_adapter: IoAdapter = None

    @abstractmethod
    def get_instance(self) -> any:
        pass

    @abstractmethod
    def create_wilchard(self, prefix: str = '/') -> str:
        pass

    @abstractmethod
    def use(self, callback: "CallableHandler", metadata: dict) -> None:
        pass

    @abstractmethod
    def get(self, route: str, callback: "CallableHandler", metadata: dict) -> None:
        pass

    @abstractmethod
    def post(self, route: str, callback: "CallableHandler", metadata: dict) -> None:
        pass

    @abstractmethod
    def ws(self, route: str, callback: WebsocketHandler, metadata: dict) -> None:
        pass

    @abstractmethod
    def mount(self, route: str, callback: MountHandler) -> None:
        pass

    @abstractmethod
    def put(self, route: str, callback: "CallableHandler", metadata: dict) -> None:
        pass

    @abstractmethod
    def delete(self, route: str, callback: "CallableHandler", metadata: dict) -> None:
        pass

    @abstractmethod
    def patch(self, route: str, callback: "CallableHandler", metadata: dict) -> None:
        pass

    @abstractmethod
    def options(self, route: str, callback: "CallableHandler", metadata: dict) -> None:
        pass

    @abstractmethod
    def head(self, route: str, callback: "CallableHandler", metadata: dict) -> None:
        pass

    @abstractmethod
    def all(self, route: str, callback: "CallableHandler", metadata: dict) -> None:
        pass

    @abstractmethod
    def engine(self, args, *kwargs) -> None:
        pass

    def use_io_adapter(self, io_adapter: IoAdapter):
        self._io_adapter = io_adapter

    def get_io_adapter(self) -> Union[IoAdapter, None]:
        return self._io_adapter

    @abstractmethod
    def enable_cors(self) -> None:
        pass

    async def on_startup(self, *_args, **_kwargs):
        for hook in self.startup_hooks:
            await hook()

    async def on_shutdown(self, *_args, **_kwargs):
        for hook in self.shutdown_hook:
            await hook()

    def on_startup_callback(self, callback: Callable):
        self.startup_hooks.append(callback)

    def on_shutdown_callback(self, callback: Callable):
        self.shutdown_hook.append(callback)

    def set(self, key: str, value: Any = None) -> None:
        SetMetadata(self.STATE_KEY, {key: value}, as_dict=True)(self)

    def get_state(self, key: str) -> Any:
        meta: dict = Reflect.get_metadata(self, self.STATE_KEY, {})
        if key in meta.keys():
            return meta[key]
        else:
            return None

    def add_global_interceptors(self, *interceptors: Union["NestipyInterceptor", Type]):
        self._global_interceptors = self._global_interceptors + list(interceptors)

    def get_global_interceptors(self):
        return self._global_interceptors

    def add_global_filters(self, *filters: Union["ExceptionFilter", Type]):
        self._global_filters = self._global_filters + list(filters)

    def get_global_filters(self):
        return self._global_filters

    def add_global_guards(self, *guards: Union["CanActivate", Type]):
        self._global_guards = self._global_guards + list(guards)

    def get_global_guards(self):
        return self._global_guards

    async def __call__(self, scope, receive, send):
        self.scope = scope
        self.receive = receive
        self.send = send
        response_sent: bool = False
        if self._io_adapter is not None:
            # Register socketio adapter
            response_sent = await self._io_adapter(scope, receive, send)
        # pass to http handler
        if not response_sent:
            await self.get_instance()(scope, receive, send)

    def create_websocket_parameter(self) -> Websocket:
        return Websocket(
            self.scope,
            self.receive,
            self.send
        )

    def create_handler_parameter(self) -> Tuple[Request, Response, NextFn]:
        req = Request(
            self.scope,
            self.receive,
            self.send
        )
        res = Response(template_engine=self.get_state(TemplateKey.MetaEngine))

        async def next_fn(error: HttpException = None):
            #  catch error
            if error is not None:
                accept = req.headers.get('accept')
                if 'application/json' in accept:
                    return await res.status(error.status_code).json({
                        "message": error.message,
                        "status": error.status_code,
                        "details": error.details
                    })
                else:
                    return await res.status(error.status_code).send(str(error))
            else:
                return await res.status(204).send("No content")

        return req, res, next_fn



from typing import Type

from nestipy.common.decorator import Injectable

from nestipy.types_ import NextFn
from nestipy.common.middleware.interface import NestipyMiddleware
from nestipy.common.http_ import Request, Response


def cors() -> Type:
    @Injectable()
    class CorsMiddleware(NestipyMiddleware):
        async def use(self, req: Request, res: Response, next_fn: NextFn):
            result = await next_fn()
            res.header('access-control-allow-origin', '*')
            res.header('access-control-allow-headers', 'content-type')
            res.header('access-control-allow-methods', 'GET, POST, PUT, DELETE, OPTIONS')
            return result

    return CorsMiddleware



from typing import Type, Union

from nestipy.metadata import SetMetadata

from .can_activate import CanActivate
from .meta import GuardKey


def UseGuards(*guards: Union[Type, CanActivate]):
    return SetMetadata(GuardKey.Meta, [g for g in guards if issubclass(g, CanActivate) or callable(g)],
                       as_list=True)



class CtxDepKey:
    Service: str = 'service'
    Request: str = 'request'
    Response: str = 'response'
    Session: str = 'session'
    Sessions: str = 'sessions'
    Cookie: str = 'cookie'
    Cookies: str = 'cookies'
    Header: str = 'header'
    Headers: str = 'headers'
    Query: str = 'query'
    Queries: str = 'queries'
    Body: str = 'body'
    Params: str = 'params'
    Param: str = 'param'
    Args: str = 'args'
    Arg: str = 'arg'
    Context: str = 'execution_context'
    Files: str = 'files'
    SocketClient: str = 'io_client'
    SocketData: str = 'io_data'

    @classmethod
    def to_list(cls) -> list:
        return [
            cls.Service, cls.Request, cls.Param,
            cls.Response, cls.Session, cls.Queries,
            cls.Query, cls.Body,
            cls.Params, cls.Args, cls.Arg,
            cls.SocketClient, cls.Cookie,
            cls.Files, cls.Cookies,
            cls.Context, cls.Sessions,
            cls.SocketData, cls.Header, cls.Headers
        ]



from typing import Type

from nestipy.metadata import ModuleMetadata, Reflect

from .metadata_creator import MetadataCreator


class ControllerMetadataCreator(MetadataCreator):
    def _type(self) -> Type["MetadataCreator"]:
        return ControllerMetadataCreator

    def _extract(self) -> list:
        return Reflect.get_metadata(self.module, ModuleMetadata.Controllers, [])



from typing import Union, Type, Callable

from nestipy.ioc import MiddlewareContainer, MiddlewareProxy


class MiddlewareConsumer:
    _module: Union[Type, object] = None

    def __init__(self, module: Union[Type, object]):
        self._module = module

    def apply(self, *middleware: Union[Type, Callable]) -> MiddlewareProxy:
        # for middleware in middlewares:
        proxy = MiddlewareProxy(*middleware)
        MiddlewareContainer.get_instance().add_singleton(proxy, self._module)
        return proxy



from .blacksheep_adapter import BlackSheepAdapter
from .fastapi_adapter import FastApiAdapter
from .http_adapter import HttpAdapter

__all__ = [
    "HttpAdapter",
    "FastApiAdapter",
    "BlackSheepAdapter"
]



from typing import Callable

from nestipy.metadata import Reflect

from .meta import TemplateKey


def Render(template: str):
    def wrapper(method: Callable):
        Reflect.set_metadata(method, TemplateKey.MetaRender, template)
        return method

    return wrapper



from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:
    from nestipy.core import ExecutionContext
    from nestipy.ioc import NestipyContainer


class RequestContextContainer:
    _instance: Union["RequestContextContainer", None] = None
    execution_context: Union["ExecutionContext", None] = None
    container: Union["NestipyContainer", None] = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(RequestContextContainer, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    @classmethod
    def set_execution_context(cls, context: "ExecutionContext"):
        ins = cls.get_instance()
        ins.execution_context = context

    @classmethod
    def set_container(cls, container: "NestipyContainer"):
        ins = cls.get_instance()
        ins.container = container

    @classmethod
    def get_instance(cls, *args, **kwargs):
        return RequestContextContainer(*args, **kwargs)

    def destroy(self):
        self._instance = None



import logging
import typing
from typing import Type

from .adapter.blacksheep_adapter import BlackSheepAdapter
from .nestipy_application import NestipyApplication, NestipyApplicationConfig


class _NestipyFactoryMeta(type):
    def __getitem__(self, item) -> "NestipyFactory":
        setattr(self, '__generic_type__', item.__name__)
        return typing.cast(NestipyFactory, self)


class NestipyFactory(metaclass=_NestipyFactoryMeta):

    @classmethod
    def create(cls, module: Type, config: NestipyApplicationConfig = None) -> NestipyApplication:
        cls._setup_log()
        if getattr(cls, '__generic_type__', None) == 'NestipyBlackSheepApplication':
            if not config:
                config = NestipyApplicationConfig(adapter=BlackSheepAdapter())
        application = NestipyApplication(config=config)
        application.init(module)
        return application

    @classmethod
    def create_micro_service(cls):
        pass

    @classmethod
    def _setup_log(cls):
        logging.addLevelName(logging.INFO, "[NESTIPY] INFO")
        logging.addLevelName(logging.ERROR, "[NESTIPY] ERROR")
        logging.addLevelName(logging.WARNING, "[NESTIPY] WARNING")
        logging.addLevelName(logging.WARN, "[NESTIPY] WARN")
        logging.addLevelName(logging.CRITICAL, "[NESTIPY] CRITICAL")



from typing import Callable, Literal

from nestipy.metadata import Reflect, RouteKey

HTTPMethod = Literal['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'HEAD', 'OPTIONS', 'ALL', 'ANY']


class Route:
    def __init__(self, path: str = '', method: list[HTTPMethod] = None, **kwargs):
        if method is None:
            method = ['GET']
        self.path = path
        self.kwargs = kwargs
        self.method = method

    def __call__(self, handler: Callable):
        # put path and kwargs in controller handler
        Reflect.set_metadata(handler, RouteKey.path, self.path)
        Reflect.set_metadata(handler, RouteKey.kwargs, self.kwargs)
        Reflect.set_metadata(handler, RouteKey.method, self.method)
        return handler


def Get(path: str = '', **kwargs):
    return Route(path=path, method=['GET'], **kwargs)


def Post(path: str = '', **kwargs):
    return Route(path=path, method=['POST'], **kwargs)


def Put(path: str = '', **kwargs):
    return Route(path=path, method=['PUT'], **kwargs)


def Patch(path: str = '', **kwargs):
    return Route(path=path, method=['PATCH'], **kwargs)


def Delete(path: str = '', **kwargs):
    return Route(path=path, method=['DELETE'], **kwargs)



"""
This module provides common functions that apply to all versions of the
OpenAPI Documentation specifications, when generating other kinds of representations
from source OAD files.
"""

from abc import ABC, abstractmethod


class DocumentsWriter(ABC):
    """
    Abstract class for types that can produce other kinds of text representations from
    source OpenAPI Documentation.
    """

    @abstractmethod
    def write(self, data, **kwargs) -> str:
        """
        Writes markdown.
        """


def is_reference(data) -> bool:
    """
    Returns a value indicating whether the given dictionary represents
    a reference.

    is_reference({"$ref": "..."}) -> True
    """
    if not isinstance(data, dict):
        return False
    return "$ref" in data


def is_object_schema(data) -> bool:
    """
    Returns a value indicating whether the given schema dictionary represents
    an object schema.

    is_reference({"type": "array", "items": {...}}) -> True
    """
    if not isinstance(data, dict):
        return False
    return data.get("type") == "object" and isinstance(data.get("properties"), dict)


def is_array_schema(data) -> bool:
    """
    Returns a value indicating whether the given schema dictionary represents
    an array schema.

    is_reference({"type": "array", "items": {...}}) -> True
    """
    if not isinstance(data, dict):
        return False
    return data.get("type") == "array" and isinstance(data.get("items"), dict)


def get_ref_type_name(reference) -> str:
    """
    Returns the type name of a reference.

    get_ref_type_name("#/components/schemas/Foo") -> "Foo"
    """
    if isinstance(reference, dict):
        reference = reference["$ref"]
    return reference.lstrip("#/").split("/")[-1]



from typing import TypeVar, Callable, Union, Type, Any

from .context_container import RequestContextContainer

TypeAnnotatedCallable = Callable[[Union[str, None], Type, RequestContextContainer], Any]


class ParamAnnotation:
    def __init__(self, callback: TypeAnnotatedCallable, key: str, token: Union[str, None] = None):
        self.key: Union[str, None] = key
        self.token: Union[str, None] = token
        self.callback = callback



def uniq(data: list) -> list:
    return list(set(data))


def uniq_list(data: list) -> list:
    uniq_data = []
    for d in data:
        if d not in uniq_data:
            uniq_data.append(d)
    return uniq_data


def deep_merge(dict_from: dict, dict_to: dict):
    keys1 = dict_to.keys()
    keys2 = dict_from.keys()
    for k in keys2:
        if k in keys1:
            dict_to_1 = dict_to[k]
            dict_from_1 = dict_from[k]
            if isinstance(dict_to_1, dict) and isinstance(dict_from_1, dict):
                dict_to_1 = deep_merge(dict_from_1, dict_to_1)
                dict_to[k] = dict_to_1
            elif isinstance(dict_to_1, list) and isinstance(dict_from_1, list, ):
                dict_to[k] = dict_to_1 + dict_from_1
            elif isinstance(dict_to_1, set, ) and isinstance(dict_from_1, set, ):
                dict_to[k] = list(dict_to_1) + list(dict_from_1)
            else:
                pass
        else:
            dict_to[k] = dict_from[k]
    return dict_to


def snakecase_to_camelcase(word: str):
    return ' '.join(word.split('_')).capitalize()


if __name__ == '__main__':
    d1 = {
        '/test': {
            'get': {
                'response': {},
                'security': {},
                'tags': [],
                'parameters': {}
            }
        }
    }

    d2 = {
        '/test': {
            'post': {
                'response': {},
                'security': {},
                'tags': [],
                'parameters': {}
            }
        }
    }
    res = deep_merge(d1, d2)

    print(res)



from abc import ABC, abstractmethod
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from nestipy.core.context.argument_host import ArgumentHost
    from nestipy.common.exception.http import HttpException


class ExceptionFilter(ABC):
    @abstractmethod
    async def catch(self, exception: "HttpException", host: "ArgumentHost") -> Any:
        pass



from typing import Callable, Awaitable, Any, Union, AsyncIterator

from nestipy.common.http_ import Request, Response, Websocket

NextFn = Callable[..., Union[Awaitable[Any]]]

CallableHandler = Callable[[Request, Response, NextFn], Union[Awaitable[Response], Response, str, dict, list]]

WebsocketHandler = Callable[[Websocket], Any]

MountHandler = Callable[[dict, Callable, Callable], Any]



from .class_ import Module, Controller, Injectable
from .method import Route, Get, Post, Patch, Put, Delete

__all__ = [
    "Module",
    "Controller",
    "Injectable",
    "Route",
    "Get",
    "Post",
    "Put",
    "Patch",
    "Delete"
]



from typing import Union, Type, TYPE_CHECKING

from nestipy.metadata import SetMetadata

from nestipy.common.decorator import Injectable
from .meta import ExceptionKey

if TYPE_CHECKING:
    from .interface import ExceptionFilter
    from nestipy.common.exception.http import HttpException


def Catch(*exceptions: Union[Type["HttpException"], "HttpException"]):
    decorator = SetMetadata(ExceptionKey.MetaType, list(exceptions), as_list=True)

    def wrapper(cls: Type["ExceptionFilter"]):
        cls = Injectable()(cls)
        return decorator(cls)

    return wrapper


def UseFilters(*filters: Union[Type["ExceptionFilter"], "ExceptionFilter"]):
    return SetMetadata(ExceptionKey.MetaFilter, list(filters), as_list=True)



from .openapi_docs.v3 import Info, OpenAPI, Components, SecurityScheme, HTTPSecurity


class DocumentBuilder:
    _info: dict = {
        'title': 'Nestipy',
        'description': '',
        'version': '1.0'
    }
    _security_schemes: dict = {}
    _schemas: dict = {}
    _instance: "DocumentBuilder" = None

    @classmethod
    def get_instance(cls):
        return DocumentBuilder()

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(DocumentBuilder, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def set_title(self, title: str):
        self._info['title'] = title
        return self

    def set_description(self, description: str):
        self._info['description'] = description
        return self

    def set_version(self, version: str):
        self._info['version'] = version
        return self

    def add_basic_auth(self):
        self._security_schemes['basicAuth'] = HTTPSecurity(
            scheme='basic',
        )
        return self

    def add_bearer_auth(self):
        self._security_schemes['bearer'] = HTTPSecurity(
            scheme='bearer',
        )
        return self

    def add_security(self, name, security: SecurityScheme):
        self._security_schemes[name] = security
        return self

    def build(self):
        return OpenAPI(
            info=Info(
                **self._info
            ),
            components=Components(
                security_schemes=self._security_schemes,
                schemas=self._schemas
            )
        )



import dataclasses
from typing import Union, Optional, Any, Type

from pydantic import TypeAdapter, BaseModel

from nestipy.metadata import SetMetadata
from .openapi_docs.v3 import RequestBody, Response, Parameter
from .openapi_docs.v3 import SecurityRequirement, MediaType, Schema


def ApiBody(body: Union[BaseModel, Type] = None, consumer: str = 'application/json'):
    content: dict = {}
    if dataclasses.is_dataclass(body):
        content: dict = TypeAdapter(body).json_schema()
    elif isinstance(body, BaseModel):
        content: dict = body.model_json_schema()
    body = RequestBody(
        required=True,
        content={
            consumer: MediaType(
                schema=Schema(**content)
            )
        },
    )
    return SetMetadata(key='__openapi__request_body', data=body)


def ApiResponse(status: int, response: Response):
    return SetMetadata(key='__openapi__responses', data={status: response}, as_dict=True)


def ApiOkResponse(description: Optional[str] = None, schema: Schema = None, example: Any = None):
    return ApiResponse(
        status=200,
        response=Response(
            description or 'Success response',
            content={
                'application/json': MediaType(
                    schema=schema or Schema(),
                    example=example
                )
            }
        )
    )


def ApiCreatedResponse(description: Optional[str] = None, schema: Schema = None, example: Any = None):
    return ApiResponse(
        status=201,
        response=Response(
            description or 'Success response',
            content={
                'application/json': MediaType(
                    schema=schema or Schema(),
                    example=example
                )
            }
        )
    )


def ApiNotFoundResponse(description: Optional[str] = None, schema: Schema = None, example: Any = None):
    return ApiResponse(
        status=404,
        response=Response(
            description or 'Not Found',
            content={
                'application/json': MediaType(
                    schema=schema or Schema(),
                    example=example
                )
            }
        )
    )


def ApiParameter(param: Parameter):
    return SetMetadata(key='__openapi__parameters', data=[param], as_list=True)


def ApiTags(tags: Union[str, list[str]]):
    if isinstance(tags, str):
        tags = [tags]
    return SetMetadata(key='__openapi__tags', data=tags, as_list=True)


def ApiId(api_id: str):
    return SetMetadata(key='__openapi__operation_id', data=api_id)


def ApiBasicAuth():
    return SetMetadata(key='__openapi__security', data=[SecurityRequirement(
        name='basicAuth',
        value=[
            'basicAuth'
        ]
    )], as_list=True)


def ApiBearerAuth():
    return SetMetadata(key='__openapi__security', data=[SecurityRequirement(
        name='bearer',
        value=[
            'bearer'
        ]
    )], as_list=True)


def ApiSecurity(security: SecurityRequirement):
    return SetMetadata(key='__openapi__security', data=[security], as_list=True)

# TODO : Add more openapi decorator



import re


class RouteParamsExtractor:
    @staticmethod
    def extract_params(route_path) -> dict:
        params = {}
        regex_pattern = r"{(\w+)(:\w+)?}"  # Pattern to match route parameters like {param_name:type}
        matches = re.findall(regex_pattern, route_path)
        for match in matches:
            param_name = match[0]
            param_type = match[1][1] if match[1] else 'str'  # Default to str if type not specified
            params[param_name] = param_type
        return params



from inspect import isfunction, getmembers
from typing import List, Type, Union

from nestipy.ioc import MiddlewareContainer
from nestipy.metadata import ModuleMetadata, Reflect, RouteKey
from openapidocs.v3 import Parameter, ParameterLocation

from nestipy.common.utils import deep_merge
from nestipy.openapi.explorer import OpenApiExplorer
from .route_extractor import RouteParamsExtractor


class RouteExplorer:
    _middleware_container: MiddlewareContainer
    _openapi_scanner = OpenApiExplorer()

    @classmethod
    def _normalize_path(cls, path: str) -> str:
        return path.strip('/')

    @classmethod
    def explore(cls, module_ref: Union[Type, object]):
        controllers = Reflect.get_metadata(module_ref, ModuleMetadata.Controllers, [])
        routes = []
        ins = cls()
        for controller in controllers:
            controller_routes = ins.explore_controller(controller)
            routes = routes + controller_routes
        return routes

    def explore_controller(self, controller: Type) -> List[dict]:

        docs = self._openapi_scanner.explore(controller)
        routes = []
        # need to extract middleware, guards, exception_handler,
        controller_path = self._normalize_path(Reflect.get_metadata(controller, RouteKey.path, ''))
        for method_name, _ in getmembers(controller, isfunction):
            if method_name.startswith("__"):
                continue
            method = getattr(controller, method_name)
            path = Reflect.get_metadata(method, RouteKey.path, None)
            if path is not None:
                method_path = self._normalize_path(path)
                path = f"/{self._normalize_path(f'{controller_path}/{method_path}')}"
                request_method = Reflect.get_metadata(method, RouteKey.method)

                # openapi docs
                params = RouteParamsExtractor.extract_params(path)
                parameters = [Parameter(name=name, in_=ParameterLocation.PATH, required=True, description='Route path',
                                        allow_empty_value=False)
                              for name in params.keys()]
                method_docs = deep_merge(self._openapi_scanner.explore(method), {'parameters': parameters})
                path_docs = deep_merge(
                    {'tags': [controller.__name__.replace('Controller', '')]},
                    deep_merge(docs, method_docs)
                )
                path_docs['tags'] = [path_docs['tags'][0]]

                # en open api

                route_info = {
                    'path': path,
                    'request_method': request_method,
                    'method_name': method_name,
                    'controller': controller,
                    'openapi': path_docs  # openapi docs
                }
                routes.append(route_info)
        return routes



import dataclasses
from typing import Literal, Annotated

from nestipy.common.decorator import Module
from nestipy.dynamic_module import ConfigurableModuleBuilder
from nestipy.ioc import Inject


@dataclasses.dataclass
class GraphqlOption:
    url: str = '/graphql'
    ide: Literal['default', 'graphiql'] = 'default'


ConfigurableModuleClass, CONFIG_MODULE_OPTION_TOKEN = ConfigurableModuleBuilder[GraphqlOption]().set_method(
    'for_root').build()


@Module()
class GraphqlModule(ConfigurableModuleClass):
    config: Annotated[GraphqlOption, Inject(CONFIG_MODULE_OPTION_TOKEN)]



from .openapi_docs.common import Serializer
from .openapi_docs.v3 import OpenAPI, Info, Response, Components, Schema, SecuritySchemeType
from .openapi_docs.v3 import PathItem, Operation, SecurityRequirement, ValueType
from .openapi_docs.v3 import Tag, HTTPSecurity, APIKeySecurity, ParameterLocation, Parameter

open_api = OpenAPI(
    info=Info(
        title='Test openapi',
        description='This is the description',
        version='v1'
    ),
    paths={
        '/test': PathItem(
            summary='Test',
            get=Operation(
                responses={
                    '200': Response(description='Response successfully')
                },
                tags=['Test'],
                operation_id='test',
                security=[
                    SecurityRequirement(
                        name='Bearer',
                        value=['bearer']
                    )
                ],
                parameters=[
                    Parameter(
                        name='test',
                        in_=ParameterLocation.PATH
                    ),
                    Parameter(
                        name='test2',
                        in_=ParameterLocation.QUERY
                    ),
                    Parameter(
                        name='test3',
                        in_=ParameterLocation.HEADER
                    ),
                    Parameter(
                        name='test3',
                        in_=ParameterLocation.HEADER
                    )
                ]
            )
        )
    },
    components=Components(
        schemas={
            'User': Schema(
                type=ValueType.OBJECT,
                properties={
                    'id': Schema(
                        type=ValueType.INTEGER,
                        example=1,
                        default=1
                    ),
                    'name': Schema(
                        type=ValueType.STRING,
                        example="name",
                        default="name"
                    )
                }
            )
        },

        security_schemes={
            'bearer': HTTPSecurity(
                scheme='bearer',
            ),
            'apiKey': APIKeySecurity(
                name='X-API',
                in_=ParameterLocation.HEADER
            ),
            'basicAuth': HTTPSecurity(
                scheme='basicAuth',
                type=SecuritySchemeType.HTTP
            ),
        }
    ),
    tags=[
        Tag(name='Test'),
        Tag(name='Test2')
    ],
)

serializer = Serializer()
if __name__ == '__main__':
    print(serializer.to_json(open_api))



from .annotation import ParamAnnotation
from .container import NestipyContainer
from .context_container import RequestContextContainer
from .dependency import Inject, Res, Req, Session, Query, Body, Arg, Context, GraphQlContext, SocketServer, \
    SocketClient, \
    SocketData, Param, Header, Cookie
from .middleware import MiddlewareProxy, MiddlewareRouteConfig, MiddlewareContainer
from .provider import ModuleProviderDict

__all__ = [
    "ParamAnnotation",
    "NestipyContainer",
    "RequestContextContainer",
    "ModuleProviderDict",
    "Inject",
    "Res",
    "Req",
    "Session",
    "Query",
    "Body",
    "Arg",
    "Context",
    "SocketServer",
    "SocketClient",
    "SocketData",
    "Param",
    "MiddlewareProxy",
    "MiddlewareRouteConfig",
    "MiddlewareContainer",
    "Header",
    "Cookie",
    "GraphQlContext"
]
